import math

from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.forms import get_slider_form
from flask_monitoringdashboard.core.plot import scatter, get_figure, get_margin, get_layout, get_average_bubble_size
from flask_monitoringdashboard.core.utils import get_endpoint_details, formatter
from flask_monitoringdashboard.database import FunctionCall, session_scope
from flask_monitoringdashboard.database.count import count_users
from flask_monitoringdashboard.database.endpoint import get_group_by_sorted
from flask_monitoringdashboard.database.function_calls import get_median
from flask_monitoringdashboard.database.versions import get_date_first_request, get_versions


@blueprint.route('/result/<end>/time_per_version_per_user', methods=['GET', 'POST'])
@secure
def result_time_per_version_per_user(end):
    with session_scope() as db_session:
        details = get_endpoint_details(db_session, end)
        form = get_slider_form(count_users(db_session, end))
    graph = get_time_per_version_per_user(end, form)
    return render_template('fmd_dashboard/graph-details.html', details=details, graph=graph, form=form)


def get_time_per_version_per_user(end, form):
    with session_scope() as db_session:
        group_by_list = get_group_by_sorted(db_session, end, form.get_slider_value())
        versions = get_versions(db_session, end)

        data = []
        for group_by in group_by_list:
            data.append(
                [get_median(db_session, end, FunctionCall.version == v, FunctionCall.group_by == group_by) for v in
                 versions])
        average = get_average_bubble_size(data)

        trace = []
        for i in range(len(group_by_list)):
            hovertext = ['Version: {}<br>Time: {}'.format(versions[j], formatter(data[i][j])) for j in
                         range(len(versions))]
            trace.append(scatter(
                x=['{}<br>{}'.format(v, get_date_first_request(db_session, v).strftime('%b %d %H:%M')) for v in
                   versions],
                hovertext=hovertext,
                y=[group_by_list[i]] * len(versions),
                name=group_by_list[i],
                mode='markers',
                marker={
                    'color': [get_color(group_by_list[i])] * len(versions),
                    'size': [math.sqrt(d) for d in data[i]],
                    'sizeref': average,
                    'sizemode': 'area'
                }
            ))

    layout = get_layout(
        height=350 + 40 * len(trace),
        title='Average execution time for every user per version',
        xaxis={'title': 'Versions', 'type': 'category'},
        yaxis={'title': 'Users', 'type': 'category', 'autorange': 'reversed'},
        margin=get_margin(b=200)
    )
    return get_figure(layout, trace)



"""
    This file contains all unit tests for the monitor-rules-table in the database. (Corresponding to the file:
    'flask_monitoringdashboard/database/tests.py')
    See __init__.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, mean, \
    EXECUTION_TIMES, NAME

NAME2 = 'main2'
SUITE = 3


class TestDBTests(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_get_tests(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_tests, Tests
        with session_scope() as db_session:
            result = get_tests(db_session)
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0].name, Tests(name=NAME).name)
            self.assertTrue(result[0].run)

    def test_add_test(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import add_test, get_tests
        with session_scope() as db_session:
            self.test_get_tests()  # checks if there is 1 test in the DB
            add_test(db_session, NAME2)
            self.assertEqual(len(get_tests(db_session)), 2)

    def test_add_or_update_test(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import add_or_update_test, get_tests
        import datetime
        with session_scope() as db_session:
            self.test_add_test()  # adds a new test
            self.assertEqual(NAME2, get_tests(db_session)[1].name)
            time = datetime.datetime.now()
            succeeded = True
            add_or_update_test(db_session, NAME2, time, succeeded)
            result = get_tests(db_session)
            self.assertEqual(len(result), 2)
            self.assertEqual(result[1].name, NAME2)
            self.assertEqual(result[1].lastRun, time)
            self.assertEqual(result[1].succeeded, succeeded)

    def test_add_test_result(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_results, add_test_result
        from flask_monitoringdashboard import config
        import datetime
        with session_scope() as db_session:
            self.assertEqual(get_results(db_session), [])
            for exec_time in EXECUTION_TIMES:
                add_test_result(db_session, NAME, exec_time, datetime.datetime.now(), config.version, SUITE, 2)
            result = get_results(db_session)
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0].name, NAME)
            self.assertEqual(result[0].count, len(EXECUTION_TIMES))
            self.assertEqual(result[0].average, mean(EXECUTION_TIMES))

    def test_get_suite_nr(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_suite_nr, add_test_result
        from flask_monitoringdashboard import config
        import datetime
        with session_scope() as db_session:
            self.assertEqual(get_suite_nr(db_session), 1)
            add_test_result(db_session, NAME, 1234, datetime.datetime.now(), config.version, SUITE, 2)
            self.assertEqual(get_suite_nr(db_session), SUITE+1)

    def test_get_results(self):
        """
            Test whether the function returns the right values.
        """
        self.test_add_test_result()  # can be replaced by test_add_test_result, since this function covers two tests

    def test_get_res_current(self):
        """
            Test whether the function returns the right values.
        """
        self.test_add_test_result()
        from flask_monitoringdashboard.database.tests import get_res_current
        from flask_monitoringdashboard import config
        with session_scope() as db_session:
            result = get_res_current(db_session, config.version)
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0].name, NAME)
            self.assertEqual(result[0].count, len(EXECUTION_TIMES))
            self.assertEqual(result[0].average, mean(EXECUTION_TIMES))

            new_version = 'new_version'
            self.assertNotEqual(new_version, config.version)
            self.assertEqual(get_res_current(db_session, new_version), [])

    def test_get_line_results(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_line_results
        from flask_monitoringdashboard import config
        with session_scope() as db_session:
            self.test_add_test_result()
            result = get_line_results(db_session)
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0].version, config.version)
            self.assertEqual(result[0].avg, mean(EXECUTION_TIMES))
            self.assertEqual(result[0].min, min(EXECUTION_TIMES))
            self.assertEqual(result[0].max, max(EXECUTION_TIMES))
            self.assertEqual(result[0].count, len(EXECUTION_TIMES))

    def test_get_suites(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_suites
        self.test_add_test_result()
        with session_scope() as db_session:
            self.assertEqual(get_suites(db_session), [(SUITE, )])

    def test_get_measurements(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_measurements
        from flask_monitoringdashboard import config
        with session_scope() as db_session:
            self.assertEqual(get_measurements(db_session, SUITE), [])
            self.test_add_test_result()
            result = get_measurements(db_session, SUITE)
            self.assertEqual(len(result), len(EXECUTION_TIMES))
            for test_run in result:
                self.assertEqual(test_run.name, NAME)
                self.assertIn(test_run.execution_time, EXECUTION_TIMES)
                self.assertEqual(test_run.version, config.version)
                self.assertEqual(test_run.suite, SUITE)

    def test_get_test_measurements(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_test_measurements
        from flask_monitoringdashboard import config
        with session_scope() as db_session:
            self.assertEqual(get_test_measurements(db_session, NAME, SUITE), [])
            self.test_add_test_result()
            result = get_test_measurements(db_session, NAME, SUITE)
            self.assertEqual(len(result), len(EXECUTION_TIMES))
            for test_run in result:
                self.assertEqual(test_run.name, NAME)
                self.assertIn(test_run.execution_time, EXECUTION_TIMES)
                self.assertEqual(test_run.version, config.version)
                self.assertEqual(test_run.suite, SUITE)



"""
    The app tracks the performance of various endpoints over time.
    To bind, use the following lines of code:
        import dashboard
        from flask import Flask
        ...
        app = Flask(__name__)
        ...
        dashboard.bind(app)

    The dashboard with the results that are collected can be found at:
        localhost:5000/dashboard
"""

import os

from flask import Blueprint

from flask_monitoringdashboard.core.config import Config

config = Config()
user_app = None


# get current location of the project
def loc():
    return os.path.abspath(os.path.dirname(__file__)) + '/'


# define the blueprint
blueprint = Blueprint('dashboard', __name__, template_folder=loc() + 'templates')


def bind(app):
    """
        Binding the app to this object should happen before importing the routing-
        methods below. Thus, the importing statement is part of this function.
        :param app: the app for which the performance has to be tracked
    """
    assert app is not None
    global user_app, blueprint
    user_app = app

    # Provide a secret-key for using WTF-forms
    if not user_app.secret_key:
        print('WARNING: You should provide a security key.')
        user_app.secret_key = 'my-secret-key'

    import os
    # Only initialize unit test logging when running on Travis.
    if '/home/travis/build/' in os.getcwd():
        print('Detected running on Travis.')
        import datetime
        from flask import request

        @user_app.after_request
        def after_request(response):
            hit_time_stamp = str(datetime.datetime.now())
            home = os.path.expanduser("~")
            log = open(home + '/endpoint_hits.log', 'a')
            log.write('"{}","{}"\n'.format(hit_time_stamp, request.endpoint))
            log.close()
            return response

    # Add all route-functions to the blueprint
    import flask_monitoringdashboard.views

    # Add wrappers to the endpoints that have to be monitored
    from flask_monitoringdashboard.core.measurement import init_measurement
    blueprint.before_app_first_request(init_measurement)

    # register the blueprint to the app
    app.register_blueprint(blueprint, url_prefix='/' + config.link)



from sqlalchemy import distinct, asc

from flask_monitoringdashboard.database import FunctionCall


def get_date_first_request(db_session, version=None):
    """
    :param version: optional: first date when this version is requested
    :return: a datetime object with the value of when the first request is made
    """
    result = db_session.query(FunctionCall.time). \
        filter((FunctionCall.version == version) | (version is None)).first()
    if result:
        return result[0]
    return None


def get_versions(db_session, end=None):
    """
    Returns a list of length 'limit' with the versions that are used in the application
    :param end: the versions that are used in a specific endpoint
    :return: a list with the versions (as a string)
    """
    result = db_session.query(distinct(FunctionCall.version)). \
        filter((FunctionCall.endpoint == end) | (end is None)). \
        order_by(asc(FunctionCall.time)).all()
    return [row[0] for row in result]



"""
    Some useful functions for setting up the testing environment, adding data, etc..
"""
import datetime
import time

from flask import Flask

NAME = 'main'
IP = '127.0.0.1'
GROUP_BY = '1'
EXECUTION_TIMES = [1000, 2000, 3000, 4000, 50000]
TIMES = [datetime.datetime.utcnow()] * 5
OUTLIER_COUNT = 3
for i in range(len(TIMES)):
    TIMES[i] -= datetime.timedelta(seconds=len(EXECUTION_TIMES)-i)
TEST_NAMES = ['test_name1', 'test_name2']


def set_test_environment():
    """ Override the config-object for a new testing environment. Module flask_monitoringdashboard must be imported locally. """
    import flask_monitoringdashboard
    flask_monitoringdashboard.config.database_name = 'sqlite:///test-database.db'


def clear_db():
    """ Drops and creates the tables in the database. Module flask_monitoringdashboard must be imported locally. """
    from flask_monitoringdashboard.database import get_tables, engine
    for table in get_tables():
        table.__table__.drop(engine)
        table.__table__.create(engine)


def add_fake_data():
    """ Adds data to the database for testing purposes. Module flask_monitoringdashboard must be imported locally. """
    from flask_monitoringdashboard.database import session_scope, FunctionCall, MonitorRule, Outlier, Tests,\
        TestsGrouped
    from flask_monitoringdashboard import config

    # Add functionCalls
    with session_scope() as db_session:
        for i in range(len(EXECUTION_TIMES)):
            call = FunctionCall(endpoint=NAME, execution_time=EXECUTION_TIMES[i], version=config.version,
                                time=TIMES[i], group_by=GROUP_BY, ip=IP)
            db_session.add(call)

    # Add MonitorRule
    with session_scope() as db_session:
        db_session.add(MonitorRule(endpoint=NAME, monitor=True, time_added=datetime.datetime.utcnow(),
                                   version_added=config.version, last_accessed=TIMES[0]))

    # Add Outliers
    with session_scope() as db_session:
        for i in range(OUTLIER_COUNT):
            db_session.add(Outlier(endpoint=NAME, cpu_percent='[%d, %d, %d, %d]' % (i, i + 1, i + 2, i + 3),
                                   execution_time = 100 * (i + 1), time = TIMES[i]))
    # Add Tests
    with session_scope() as db_session:
        db_session.add(Tests(name=NAME, succeeded=True))

    # Add TestsGrouped
    with session_scope() as db_session:
        for test_name in TEST_NAMES:
            db_session.add(TestsGrouped(endpoint=NAME, test_name=test_name))


def get_test_app():
    """
    :return: Flask Test Application with the right settings
    """
    import flask_monitoringdashboard
    user_app = Flask(__name__)
    user_app.config['SECRET_KEY'] = flask_monitoringdashboard.config.security_token
    user_app.testing = True
    flask_monitoringdashboard.user_app = user_app
    user_app.config['WTF_CSRF_ENABLED'] = False
    user_app.config['WTF_CSRF_METHODS'] = []
    flask_monitoringdashboard.config.get_group_by = lambda: '12345'
    flask_monitoringdashboard.bind(app=user_app)
    return user_app


def login(test_app):
    """
    Used for setting the sessions, such that you have a fake login to the flask_monitoringdashboard.
    :param test_app:
    """
    from flask_monitoringdashboard import config
    with test_app.session_transaction() as sess:
        sess[config.link + '_logged_in'] = True
        sess[config.link + '_admin'] = True


def mean(numbers):
    return float(sum(numbers)) / max(len(numbers), 1)


def test_admin_secure(test_case, page):
    """
    Test whether the page is only accessible with admin credentials.
    :param test_case: test class, must be an instance of unittest.TestCase
    :param page: str with the page of the flask_monitoringdashboard
    """
    with test_case.app.test_client() as c:
        test_case.assertEqual(302, c.get('dashboard/{}'.format(page)).status_code)
        login(c)
        test_case.assertEqual(200, c.get('dashboard/{}'.format(page)).status_code)


import datetime

import plotly.graph_objs as go
from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.forms import get_daterange_form
from flask_monitoringdashboard.core.plot import get_layout, get_figure, heatmap as plot_heatmap
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.endpoint import get_num_requests


TITLE = 'Heatmap of the number of requests'


@blueprint.route('/measurements/heatmap', methods=['GET', 'POST'])
@secure
def heatmap():
    form = get_daterange_form()
    return render_template('fmd_dashboard/graph.html', form=form, graph=get_heatmap(form), title=TITLE)


def get_heatmap(form, end=None, title=None):
    """
    Return HTML string for generating a Heatmap.
    :param form: A SelectDateRangeForm, which is used to filter the selection
    :param end: optionally, filter the data on a specific endpoint
    :return: HTML code with the graph
    """
    # list of hours: 0:00 - 23:00
    hours = ['0{}:00'.format(h) for h in range(0, 10)] + ['{}:00'.format(h) for h in range(10, 24)]
    days = form.get_days()

    # create empty 2D-list: [hour][day]
    heatmap_data = []
    for i in range(len(hours)):
        heatmap_data.append([])
        [heatmap_data[i].append(0) for _ in days]

    # add data from database to heatmap_data
    with session_scope() as db_session:
        for d in get_num_requests(db_session, end, form.start_date.data, form.end_date.data):
            parsed_time = datetime.datetime.strptime(d.newTime, '%Y-%m-%d %H:%M:%S')
            day_index = (parsed_time - datetime.datetime.combine(form.start_date.data, datetime.time(0, 0, 0, 0))).days
            hour_index = int(parsed_time.strftime('%H'))
            heatmap_data[hour_index][day_index] = d.count

    if end:
        if not title:
            title = ''
        title += ' for endpoint: {}'.format(end)

    layout = get_layout(
        title=title,
        xaxis=go.XAxis(range=[(form.start_date.data - datetime.timedelta(days=1, hours=6)).
                       strftime('%Y-%m-%d 12:00:00'), form.end_date.data.strftime('%Y-%m-%d 12:00:00')]),
        yaxis={'autorange': 'reversed'}
    )
    return get_figure(layout, [plot_heatmap(x=days, y=hours, z=heatmap_data)])



def get_rules():
    """
    :return: A list of the current rules in the attached Flask app
    """
    from flask_monitoringdashboard import config, user_app

    rules = user_app.url_map.iter_rules()
    return [r for r in rules if not r.rule.startswith('/' + config.link)
            and not r.rule.startswith('/static-' + config.link)]



import time
import unittest


class TestOutlier(unittest.TestCase):

    def test_stackinfo(self):
        """
            One test for all outlier functionality
        """
        from flask_monitoringdashboard.core.outlier import StackInfo
        stack_info = StackInfo(average=1)
        time.sleep(0.01)
        stack_info.stop()
        self.assertIsNotNone(stack_info.stacktrace)
        self.assertIsNotNone(stack_info.cpu_percent)
        self.assertIsNotNone(stack_info.memory)



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.forms import get_slider_form
from flask_monitoringdashboard.core.plot import get_layout, get_figure, boxplot
from flask_monitoringdashboard.core.utils import get_endpoint_details
from flask_monitoringdashboard.database import FunctionCall, session_scope
from flask_monitoringdashboard.database.count import count_users
from flask_monitoringdashboard.database.endpoint import get_all_measurement_per_column, get_group_by_sorted


@blueprint.route('/result/<end>/time_per_user', methods=['GET', 'POST'])
@secure
def result_time_per_user(end):
    with session_scope() as db_session:
        details = get_endpoint_details(db_session, end)
        form = get_slider_form(count_users(db_session, end))
    graph = get_time_per_user(end, form)

    return render_template('fmd_dashboard/graph-details.html', details=details, graph=graph, form=form)


def get_time_per_user(end, form):
    """
    Return an HTML box plot with a specific number of
    :param end: get the data for this endpoint only
    :param form: instance of SliderForm
    :return:
    """
    data = []
    with session_scope() as db_session:
        for group_by in get_group_by_sorted(db_session, end, form.get_slider_value()):
            values = [str(c.execution_time) for c in
                      get_all_measurement_per_column(db_session, endpoint=end, column=FunctionCall.group_by,
                                                     value=group_by)]
            data.append(boxplot(values, name=str(group_by)))

    layout = get_layout(
        height=350 + 40 * len(data),
        title='Execution time for every user for endpoint: ' + end,
        xaxis={'title': 'Execution time (ms)'},
        yaxis={'type': 'category', 'title': 'User'}
    )
    return get_figure(layout, data)



from sqlalchemy import func, distinct

from flask_monitoringdashboard.database import FunctionCall, Outlier


def count_rows(db_session, column, *criterion):
    """
    Count the number of rows of a specified column
    :param column: column to count
    :param criterion: where-clause of the query
    :return: number of rows
    """
    result = db_session.query(func.count(distinct(column))).filter(*criterion).first()
    if result:
        return result[0]
    return 0


def count_users(db_session, endpoint):
    """
    :param endpoint: filter on this endpoint
    :return: The number of distinct users that have requested this endpoint
    """
    return count_rows(db_session, FunctionCall.group_by, FunctionCall.endpoint == endpoint)


def count_ip(db_session, endpoint):
    """
    :param endpoint: filter on this endpoint
    :return: The number of distinct users that have requested this endpoint
    """
    return count_rows(db_session, FunctionCall.ip, FunctionCall.endpoint == endpoint)


def count_versions(db_session, endpoint):
    """
    :param endpoint: filter on this endpoint
    :return: The number of distinct versions that are used for this endpoint
    """
    return count_rows(db_session, FunctionCall.version, FunctionCall.endpoint == endpoint)


def count_requests(db_session, endpoint, *where):
    """ Return the number of hits for a specific endpoint (possible with more filter arguments).
    :param endpoint: name of the endpoint
    :param where: additional arguments
    """
    return count_rows(db_session, FunctionCall.id, FunctionCall.endpoint == endpoint, *where)


def count_outliers(db_session, endpoint):
    """
    :return: An integer with the number of rows in the Outlier-table.
    """
    return count_rows(db_session, Outlier.id, Outlier.endpoint == endpoint)



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.plot import get_layout, get_figure, get_margin, heatmap
from flask_monitoringdashboard.database import FunctionCall, session_scope
from flask_monitoringdashboard.database.count import count_requests
from flask_monitoringdashboard.database.function_calls import get_endpoints
from flask_monitoringdashboard.database.versions import get_versions

TITLE = 'Distribution of the load per endpoint per version'


@blueprint.route('/measurements/version_usage')
@secure
def version_usage():
    return render_template('fmd_dashboard/graph.html', graph=get_version_usage(), title=TITLE)


def get_version_usage():
    """
    Used for getting a Heatmap with an overview of which endpoints are used in which versions
    :return:
    """
    with session_scope() as db_session:
        endpoints = get_endpoints(db_session)
        versions = get_versions(db_session)

        hits = [[count_requests(db_session, e, FunctionCall.version == v) for v in versions] for e in endpoints]

        for i in range(len(versions)):  # compute the total number of hits in a specific version
            total_hits = max(1, sum([column[i] for column in hits]))

            for j in range(len(endpoints)):  # compute distribution
                hits[j][i] = hits[j][i] * 100 / total_hits

    layout = get_layout(
        xaxis={'title': 'Versions', 'type': 'category'},
        yaxis={'type': 'category', 'autorange': 'reversed'},
        margin=get_margin()
    )

    trace = heatmap(
        z=hits,
        x=versions,
        y=['{} '.format(e) for e in endpoints],
        colorbar={
            'titleside': 'top',
            'tickmode': 'array',
        }
    )
    return get_figure(layout=layout, data=[trace])



"""
    Contains all functions that are used to track the performance of the flask-application.
    See init_measurement() for more detailed info.
"""
import datetime
import time
import traceback
from functools import wraps

from flask import request

from flask_monitoringdashboard import config
from flask_monitoringdashboard.core.outlier import StackInfo
from flask_monitoringdashboard.core.rules import get_rules
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.endpoint import update_last_accessed
from flask_monitoringdashboard.database.function_calls import add_function_call
from flask_monitoringdashboard.database.monitor_rules import get_monitor_rules
from flask_monitoringdashboard.database.outlier import add_outlier

# count and sum are dicts and used for calculating the averages
endpoint_count = {}
endpoint_sum = {}

MIN_NUM_REQUESTS = 10


def init_measurement():
    """
    This should be added to the list of functions that are executed before the first request.
    This function is used in the config-method in __init__ of this folder
    It adds wrappers to the endpoints for tracking their performance and last access times.
    """
    from flask_monitoringdashboard import user_app
    with session_scope() as db_session:
        for rule in get_monitor_rules(db_session):
            # add a wrapper for every endpoint
            if rule.endpoint in user_app.view_functions:
                user_app.view_functions[rule.endpoint] = track_performance(user_app.view_functions[rule.endpoint],
                                                                           endpoint=rule.endpoint)

    for rule in get_rules():
        user_app.view_functions[rule.endpoint] = track_last_accessed(user_app.view_functions[rule.endpoint],
                                                                     endpoint=rule.endpoint)


def track_performance(func, endpoint):
    """
    Measure the execution time of a function and store result in the database
    :param func: the function to be measured
    :param endpoint: the name of the endpoint
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            # compute average
            average = get_average(endpoint)

            stack_info = None

            if average:
                average *= config.outlier_detection_constant

                # start a thread to log the stacktrace after 'average' ms
                stack_info = StackInfo(average)

            time1 = time.time()
            result = func(*args, **kwargs)

            if stack_info:
                stack_info.stop()

            time2 = time.time()
            t = (time2 - time1) * 1000
            with session_scope() as db_session:
                add_function_call(db_session, execution_time=t, endpoint=endpoint, ip=request.environ['REMOTE_ADDR'])

            # outlier detection
            endpoint_count[endpoint] = endpoint_count.get(endpoint, 0) + 1
            endpoint_sum[endpoint] = endpoint_sum.get(endpoint, 0) + t

            if stack_info:
                with session_scope() as db_session:
                    add_outlier(db_session, endpoint, t, stack_info, request)

            return result
        except:
            traceback.print_exc()
            # Execute the endpoint that was called, even if the tracking fails.
            return func(*args, **kwargs)

    wrapper.original = func

    return wrapper


def track_last_accessed(func, endpoint):
    """
    Keep track of the last access time of the endpoints. 
    :param func: the function to be measured
    :param endpoint: the name of the endpoint
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            with session_scope() as db_session:
                update_last_accessed(db_session, endpoint=endpoint, value=datetime.datetime.utcnow())
        except:
            traceback.print_exc()

        # Execute the endpoint that was called, even if the tracking fails.
        return func(*args, **kwargs)

    return wrapper


def get_average(endpoint):
    if not config.outliers_enabled:
        return None

    if endpoint in endpoint_count:
        if endpoint_count[endpoint] < MIN_NUM_REQUESTS:
            return None
    else:
        # initialize endpoint
        endpoint_count[endpoint] = 0
        endpoint_sum[endpoint] = 0
        return None
    return endpoint_sum[endpoint] / endpoint_count[endpoint]



"""
    Creates the database. 
    For information about how to access the database via a session-variable, see: session_scope() 
"""
import datetime
from contextlib import contextmanager

from sqlalchemy import Column, Integer, String, DateTime, create_engine, Float, Boolean, TEXT, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from flask_monitoringdashboard import config
from flask_monitoringdashboard.core.group_by import get_group_by

Base = declarative_base()


class MonitorRule(Base):
    """ Table for storing which endpoints to monitor. """
    __tablename__ = 'rules'
    # endpoint must be unique and acts as a primary key
    endpoint = Column(String(250), primary_key=True)
    # boolean to determine whether the endpoint should be monitored?
    monitor = Column(Boolean, default=False)
    # the time and version on which the endpoint is added
    time_added = Column(DateTime)
    version_added = Column(String(100), default=config.version)
    # the timestamp of the last access time
    last_accessed = Column(DateTime)


class Tests(Base):
    """ Table for storing which tests to run. """
    __tablename__ = 'tests'
    # name must be unique and acts as a primary key
    name = Column(String(250), primary_key=True)
    # boolean to determine whether the test should be run
    run = Column(Boolean, default=True)
    # the timestamp of the last time the test was run
    lastRun = Column(DateTime)
    # whether the test succeeded
    succeeded = Column(Boolean)


class TestRun(Base):
    """ Table for storing test results. """
    __tablename__ = 'testRun'
    # name of executed test
    name = Column(String(250), primary_key=True)
    # execution_time in ms
    execution_time = Column(Float, primary_key=True)
    # time of adding the result to the database
    time = Column(DateTime, primary_key=True)
    # version of the website at the moment of adding the result to the database
    version = Column(String(100), nullable=False)
    # number of the test suite execution
    suite = Column(Integer)
    # number describing the i-th run of the test within the suite
    run = Column(Integer)


class FunctionCall(Base):
    """ Table for storing measurements of function calls. """
    __tablename__ = 'functionCalls'
    id = Column(Integer, primary_key=True, autoincrement=True)
    endpoint = Column(String(250), nullable=False)
    # execution_time in ms
    execution_time = Column(Float, nullable=False)
    # time of adding the result to the database
    time = Column(DateTime, default=datetime.datetime.utcnow)
    # version of the website at the moment of adding the result to the database
    version = Column(String(100), nullable=False)
    # which user is calling the function
    group_by = Column(String(100), default=get_group_by)
    # ip address of remote user
    ip = Column(String(25), nullable=False)


class Outlier(Base):
    """ Table for storing information about outliers. """
    __tablename__ = 'outliers'
    id = Column(Integer, primary_key=True, autoincrement=True)
    endpoint = Column(String(250), nullable=False)

    # request-values, GET, POST, PUT
    request_values = Column(TEXT)
    # request headers
    request_headers = Column(TEXT)
    # request environment
    request_environment = Column(TEXT)
    # request url
    request_url = Column(String(1000))

    # cpu_percent in use
    cpu_percent = Column(String(100))
    # memory
    memory = Column(TEXT)

    # stacktrace
    stacktrace = Column(TEXT)

    # execution_time in ms
    execution_time = Column(Float, nullable=False)
    # time of adding the result to the database
    time = Column(DateTime)


class TestsGrouped(Base):
    """ Table for storing grouped tests on endpoints. """
    __tablename__ = 'testsGrouped'
    # Name of the endpoint
    endpoint = Column(String(250), primary_key=True)
    # Name of the unit test
    test_name = Column(String(250), primary_key=True)


# define the database
engine = create_engine(config.database_name)

# creates all tables in the database
Base.metadata.create_all(engine)
Base.metadata.bind = engine
DBSession = sessionmaker(bind=engine)


@contextmanager
def session_scope():
    """
    When accessing the database, use the following syntax:
        with session_scope() as db_session:
            db_session.query(...)
            
    :return: the session for accessing the database
    """
    session = DBSession()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


def get_tables():
    return [MonitorRule, Tests, TestRun, FunctionCall, Outlier, TestsGrouped]


from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.forms import get_daterange_form
from flask_monitoringdashboard.core.plot import barplot, get_figure, get_layout
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.function_calls import get_requests_per_day, get_endpoints

TITLE = 'Requests per endpoint per day'


@blueprint.route('/measurements/requests', methods=['GET', 'POST'])
@secure
def page_number_of_requests_per_endpoint():
    form = get_daterange_form(num_days=10)
    return render_template('fmd_dashboard/graph.html', form=form, graph=get_stacked_bar(form), title=TITLE)


def get_stacked_bar(form):
    """
    Returns a horizontal boxplot with the number of requests per day.
    :param form: must be the form that is obtained by get_daterange_form
    :return:
    """
    days = form.get_days()
    with session_scope() as db_session:
        data = [barplot(x=get_requests_per_day(db_session, end, days), y=days, name=end) for end in
                get_endpoints(db_session)]

    layout = get_layout(
        barmode='stack',
        height=350 + 40 * len(days),
        showlegend=True,
        xaxis={'title': 'Number of requests'},
        yaxis={'type': 'category', 'autorange': 'reversed'}
    )

    return get_figure(layout=layout, data=data)



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.forms import get_daterange_form
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.views.dashboard.heatmap import get_heatmap
from flask_monitoringdashboard.core.utils import get_endpoint_details


@blueprint.route('/result/<end>/heatmap', methods=['GET', 'POST'])
@secure
def result_heatmap(end):
    form = get_daterange_form()
    with session_scope() as db_session:
        details = get_endpoint_details(db_session, end)
    return render_template('fmd_dashboard/graph-details.html', form=form, details=details, graph=get_heatmap(form, end))



from flask import request
from flask_wtf import FlaskForm
from wtforms import SubmitField
from wtforms.fields.html5 import IntegerRangeField


class SliderForm(FlaskForm):
    """
        Class for generating a slider that can be used to reduce the graph.
    """
    slider = IntegerRangeField()
    submit = SubmitField('Submit')
    type = 'SliderForm'

    def get_slider_value(self):
        """
        :return: the value from the slider
        """
        if 'slider' in request.form:
            return request.form['slider']
        return self.start_value


def get_slider_form(slider_max=100):
    """
    Return a SliderForm with the range from 0 to slider_max
    :param slider_max: maximum value for the slider
    :return: a SliderForm with the range (0 ... slider_max)
    """
    form = SliderForm(request.form)
    form.min_value = 1
    form.max_value = slider_max
    if 'slider' in request.form:
        form.start_value = request.form['slider']
    else:
        form.start_value = min(max(form.min_value, form.min_value + (slider_max - form.min_value) // 2), form.max_value)
    return form



import datetime

from flask import json
from sqlalchemy import desc

from flask_monitoringdashboard.database import Outlier


def add_outlier(db_session, endpoint, execution_time, stack_info, request):
    """ Collects information (request-parameters, memory, stacktrace) about the request and adds it in the database."""
    outlier = Outlier(endpoint=endpoint, request_values=json.dumps(request.values),
                      request_headers=str(request.headers), request_environment=str(request.environ),
                      request_url=str(request.url), cpu_percent=stack_info.cpu_percent,
                      memory=stack_info.memory, stacktrace=stack_info.stacktrace,
                      execution_time=execution_time, time=datetime.datetime.now())
    db_session.add(outlier)


def get_outliers_sorted(db_session, endpoint, sort_column, offset, per_page):
    """
    :param endpoint: only get outliers from this endpoint
    :param sort_column: column used for sorting the result
    :param offset: number of items to skip
    :param per_page: number of items to return
    :return: a list of all outliers of a specific endpoint. The list is sorted based on the column that is given.
    """
    result = db_session.query(Outlier).filter(Outlier.endpoint == endpoint).order_by(desc(sort_column)). \
        offset(offset).limit(per_page).all()
    db_session.expunge_all()
    return result


def get_outliers_cpus(db_session, endpoint):
    """
    :param db_session: the session containing the query
    :param endpoint: only get outliers from this endpoint
    :return: a list of all cpu percentages for outliers of a specific endpoint
    """
    result = db_session.query(Outlier.cpu_percent).filter(Outlier.endpoint == endpoint).all()
    return result


def delete_outliers_without_stacktrace(db_session):
    """
        Remove the outliers which don't have a stacktrace.
        This is possibly due to an error in the outlier functionality
    """
    db_session.query(Outlier).filter(Outlier.stacktrace == '').delete()



import datetime

from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.auth import secure, is_admin
from flask_monitoringdashboard.database import FunctionCall, session_scope
from flask_monitoringdashboard.database.endpoint import get_last_accessed_times
from flask_monitoringdashboard.database.function_calls import get_median, get_endpoints
from flask_monitoringdashboard.database.count import count_requests


@blueprint.route('/measurements/overview')
@secure
def overview():
    week_ago = datetime.datetime.utcnow() - datetime.timedelta(days=7)
    today = datetime.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)

    result = []
    with session_scope() as db_session:
        for endpoint in get_endpoints(db_session):
            result.append({
                'name': endpoint,
                'color': get_color(endpoint),
                'hits-today': count_requests(db_session, endpoint, FunctionCall.time > today),
                'hits-week': count_requests(db_session, endpoint, FunctionCall.time > week_ago),
                'hits-overall': count_requests(db_session, endpoint),
                'median-today': get_median(db_session, endpoint, FunctionCall.time > today),
                'median-week': get_median(db_session, endpoint, FunctionCall.time > week_ago),
                'median-overall': get_median(db_session, endpoint),
                'last-accessed': get_last_accessed_times(db_session, endpoint)
            })

    return render_template('fmd_dashboard/overview.html', result=result, is_admin=is_admin(),
                           title='Dashboard Overview')



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.plot import boxplot, get_figure, get_layout, get_margin
from flask_monitoringdashboard.core.utils import get_endpoint_details
from flask_monitoringdashboard.database import FunctionCall, session_scope
from flask_monitoringdashboard.database.endpoint import get_all_measurement_per_column
from flask_monitoringdashboard.database.versions import get_date_first_request, get_versions


@blueprint.route('/result/<end>/time_per_version', methods=['GET', 'POST'])
@secure
def result_time_per_version(end):
    with session_scope() as db_session:
        details = get_endpoint_details(db_session, end)
    graph = get_time_per_version(end)
    return render_template('fmd_dashboard/graph-details.html', details=details, graph=graph)


def get_time_per_version(end):
    data = []
    with session_scope() as db_session:
        versions = get_versions(db_session, end=end)
        for version in versions:
            values = [str(c.execution_time) for c in
                      get_all_measurement_per_column(db_session, endpoint=end, column=FunctionCall.version,
                                                     value=version)]
            name = "{} {}".format(version, get_date_first_request(db_session, version).strftime("%b %d %H:%M"))
            data.append(boxplot(values=values, name=name))

    layout = get_layout(
        height=350 + 40 * len(versions),
        title='Execution time for every version',
        xaxis={'title': 'Execution time (ms)'},
        yaxis={'type': 'category', 'title': 'Version', 'autorange': 'reversed'},
        margin=get_margin()
    )
    return get_figure(layout=layout, data=data)



from flask import request, render_template

from flask_monitoringdashboard import blueprint, user_app
from flask_monitoringdashboard.core.auth import admin_secure
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.forms import MonitorDashboard
from flask_monitoringdashboard.core.measurement import track_performance
from flask_monitoringdashboard.core.rules import get_rules
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.endpoint import get_monitor_rule, update_monitor_rule, get_last_accessed_times
from flask_monitoringdashboard.database.monitor_rules import reset_monitor_endpoints


@blueprint.route('/rules', methods=['GET', 'POST'])
@admin_secure
def rules():
    """
    Renders a table with all rules from the user_app. The fmd_dashboard rules are excluded
    In case of the POST request, the data from the form is validated and processed, such that the required rules are
    monitored
    :return:
    """
    form = MonitorDashboard(request.form)

    with session_scope() as db_session:

        if request.method == 'POST' and form.validate():
            # Remove the monitor endpoints from the database
            reset_monitor_endpoints(db_session)

            for rule in get_rules():
                # Remove existing wrappers
                original = getattr(user_app.view_functions[rule.endpoint], 'original', None)
                if original:
                    user_app.view_functions[rule.endpoint] = original

            # request.form only contains checkboxes that are checked
            for data in request.form:
                if data.startswith('checkbox-'):
                    endpoint = data.rsplit('-', 1)[1]
                    update_monitor_rule(db_session, endpoint, value=True)
                    rule = get_monitor_rule(db_session, endpoint)
                    # Add wrappers to the existing functions
                    user_app.view_functions[rule.endpoint] = track_performance(user_app.view_functions[rule.endpoint],
                                                                               rule.endpoint)
        all_rules = []
        for rule in get_rules():
            all_rules.append({
                'color': get_color(rule.endpoint),
                'rule': rule.rule,
                'endpoint': rule.endpoint,
                'methods': rule.methods,
                'last_accessed': get_last_accessed_times(db_session, rule.endpoint),
                'monitor': get_monitor_rule(db_session, rule.endpoint).monitor
            })

    return render_template('fmd_rules.html', rules=all_rules, form=form)



"""
Contains all functions that operate on the testsGrouped table
"""
from flask_monitoringdashboard.database import TestsGrouped


def reset_tests_grouped(db_session):
    """ Resets the testsGrouped table of the database. """
    db_session.query(TestsGrouped).delete()


def add_tests_grouped(db_session, json):
    """ Adds endpoint - unit tests combinations to the database. """
    for combination in json:
        db_session.add(TestsGrouped(endpoint=combination['endpoint'], test_name=combination['test_name']))


def get_tests_grouped(db_session):
    """ Return all existing endpoint - unit tests combinations. """
    result = db_session.query(TestsGrouped).all()
    db_session.expunge_all()
    return result



"""
    Contains all endpoints that can be found in the Dashboard-menu on the left.
    The endpoints are the following:
    - Overview: shows a quick overview of every endpoint.
    - Heatmap: shows a heatmap of the number of requests.
    - Version usage: shows how much a certain version is used.
    - Requests per endpoint: shows how much requests a certain endpoint received.
    - Time per version: Shows the overall execution time for all versions.
    - Time per endpoint: Shows the overall execution time for all endpoints.
"""

from flask_monitoringdashboard.views.dashboard.endpoints import page_boxplot_per_endpoint
from flask_monitoringdashboard.views.dashboard.heatmap import heatmap
from flask_monitoringdashboard.views.dashboard.overview import overview
from flask_monitoringdashboard.views.dashboard.requests import page_number_of_requests_per_endpoint
from flask_monitoringdashboard.views.dashboard.version_usage import version_usage
from flask_monitoringdashboard.views.dashboard.versions import page_boxplot_per_version



from flask import render_template
from flask_paginate import get_page_args, Pagination

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.utils import get_endpoint_details, get_mean_cpu
from flask_monitoringdashboard.database import Outlier, session_scope
from flask_monitoringdashboard.database.count import count_outliers
from flask_monitoringdashboard.database.outlier import get_outliers_sorted, delete_outliers_without_stacktrace, \
    get_outliers_cpus

OUTLIERS_PER_PAGE = 10


@blueprint.route('/result/<end>/outliers')
@secure
def result_outliers(end):
    with session_scope() as db_session:
        details = get_endpoint_details(db_session, end)
        delete_outliers_without_stacktrace(db_session)
        page, per_page, offset = get_page_args(page_parameter='page', per_page_parameter='per_page')
        table = get_outliers_sorted(db_session, end, Outlier.execution_time, offset, per_page)
        all_cpus = get_outliers_cpus(db_session, end)
        mean = get_mean_cpu(all_cpus)
        pagination = Pagination(page=page, per_page=per_page, total=count_outliers(db_session, end), format_number=True,
                                css_framework='bootstrap4', format_total=True, record_name='outliers')

    return render_template('fmd_dashboard/outliers.html', details=details, table=table, pagination=pagination, mean=mean)



import unittest

from flask import Flask

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.count import count_requests
from flask_monitoringdashboard.database.endpoint import get_last_accessed_times, get_monitor_rule
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data

NAME = 'func'


class TestMeasurement(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = Flask(__name__)

    @staticmethod
    def func():
        """ function for testing the performance """
        return 'Hello World!'

    def test_get_group_by(self):
        """
            Test whether the group_by returns the right result
        """
        from flask_monitoringdashboard import config
        from flask_monitoringdashboard.core.group_by import get_group_by
        config.group_by = lambda: 3
        self.assertEqual(get_group_by(), '3')

        config.group_by = (lambda: 'User', lambda: 3.0)
        self.assertEqual(get_group_by(), '(User,3.0)')

    def test_track_performance(self):
        """
            Test whether the track_performance works
        """
        from flask_monitoringdashboard.core.measurement import track_performance

        with self.app.test_request_context(environ_base={'REMOTE_ADDR': '127.0.0.1'}):
            self.func = track_performance(self.func, NAME)
            self.func()
            with session_scope() as db_session:
                self.assertEqual(count_requests(db_session, NAME), 1)

    def test_last_accessed(self):
        """
            Test whether the last_accessed is stored in the database
        """
        from flask_monitoringdashboard.core.measurement import track_last_accessed

        with session_scope() as db_session:
            get_monitor_rule(db_session, NAME)
            func = track_last_accessed(self.func, NAME)

        with session_scope() as db_session:
            self.assertIsNone(get_last_accessed_times(db_session, NAME))
            func()
            self.assertIsNotNone(get_last_accessed_times(db_session, NAME))

    def test_get_average(self):
        """
            Test get_average
        """
        from flask_monitoringdashboard.core.measurement import track_performance, get_average, MIN_NUM_REQUESTS
        self.assertIsNone(get_average(NAME))

        with self.app.test_request_context(environ_base={'REMOTE_ADDR': '127.0.0.1'}):
            self.func = track_performance(self.func, NAME)
            for i in range(MIN_NUM_REQUESTS):
                self.func()

        self.assertIsNotNone(get_average(NAME))



"""
Contains all functions that returns results of all monitor_rules
"""
from flask_monitoringdashboard.database import MonitorRule


def get_monitor_rules(db_session):
    """ Return all monitor rules that are currently being monitored"""
    result = db_session.query(MonitorRule).filter(MonitorRule.monitor).all()
    db_session.expunge_all()
    return result


def get_monitor_names(db_session):
    """ Return all names of monitor rules that are currently being monitored"""
    result = db_session.query(MonitorRule.endpoint).filter(MonitorRule.monitor).all()
    db_session.expunge_all()
    return result


def reset_monitor_endpoints(db_session):
    """ Update all monitor rules in the database and set them to false. """
    db_session.query(MonitorRule).update({MonitorRule.monitor: False})


def get_monitor_data(db_session):
    """
    Returns all data in the rules-table. This table contains which endpoints are being
    monitored and which are not.
    :return: all data from the database in the rules-table.
    """
    result = db_session.query(MonitorRule.endpoint, MonitorRule.last_accessed, MonitorRule.monitor,
                              MonitorRule.time_added, MonitorRule.version_added).all()
    db_session.expunge_all()
    return result



import plotly
import plotly.graph_objs as go
from flask import session, render_template

from flask_monitoringdashboard import blueprint, config
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.monitor_rules import get_monitor_names
from flask_monitoringdashboard.database.tests import get_res_current, get_measurements
from flask_monitoringdashboard.database.tests import get_tests, get_results, get_suites, get_test_measurements
from flask_monitoringdashboard.database.tests_grouped import get_tests_grouped


@blueprint.route('/testmonitor/<test>')
@secure
def test_result(test):
    return render_template('fmd_testmonitor/testresult.html', link=config.link, session=session, name=test,
                           boxplot=get_boxplot(test))


@blueprint.route('/testmonitor')
@secure
def testmonitor():
    with session_scope() as db_session:
        endp_names = [name.endpoint for name in get_monitor_names(db_session)]

        tests = get_tests_grouped(db_session)
        grouped = {}
        cols = {}
        for t in tests:
            if t.endpoint not in grouped:
                grouped[t.endpoint] = []
                cols[t.endpoint] = get_color(t.endpoint)
            if t.test_name not in grouped[t.endpoint]:
                grouped[t.endpoint].append(t.test_name)

        return render_template('fmd_testmonitor/testmonitor.html', tests=get_tests(db_session), endpoints=endp_names,
                               results=get_results(db_session), groups=grouped, colors=cols,
                               res_current_version=get_res_current(db_session, config.version),
                               boxplot=get_boxplot(None))


def get_boxplot(test):
    data = []
    with session_scope() as db_session:
        suites = get_suites(db_session)
    if not suites:
        return None
    for s in suites:
        if test:
            values = [str(c.execution_time) for c in get_test_measurements(db_session, name=test, suite=s.suite)]
        else:
            values = [str(c.execution_time) for c in get_measurements(db_session, suite=s.suite)]

        data.append(go.Box(
            x=values,
            name="{0} ({1})".format(s.suite, len(values))))

    layout = go.Layout(
        autosize=True,
        height=350 + 40 * len(suites),
        plot_bgcolor='rgba(249,249,249,1)',
        showlegend=False,
        title='Execution times for every Travis build',
        xaxis=dict(title='Execution time (ms)'),
        yaxis=dict(
            title='Build (measurements)',
            autorange='reversed'
        )
    )

    return plotly.offline.plot(go.Figure(data=data, layout=layout), output_type='div', show_link=False)



"""
Contains all functions that returns results of all tests
"""
from sqlalchemy import func, desc

from flask_monitoringdashboard.database import Tests, TestRun


def get_tests(db_session):
    """ Return all existing tests. """
    result = db_session.query(Tests).all()
    db_session.expunge_all()
    return result


def add_test(db_session, name):
    """ Add a newly found test to the database. """
    db_session.add(Tests(name=name))


def add_or_update_test(db_session, name, last_run, succeeded):
    """ Updates values of a test. """
    if db_session.query(Tests).filter(Tests.name == name).count():
        db_session.query(Tests).filter(Tests.name == name). \
            update({Tests.lastRun: last_run, Tests.succeeded: succeeded})
    else:
        db_session.add(Tests(name=name, lastRun=last_run, succeeded=succeeded))


def add_test_result(db_session, name, exec_time, time, version, suite, iter):
    """ Add a test result to the database. """
    db_session.add(
        TestRun(name=name, execution_time=exec_time, time=time, version=version, suite=suite, run=iter))


def get_suite_nr(db_session):
    """ Retrieves the number of the next suite to run. """
    result = db_session.query(func.max(TestRun.suite).label('nr')).one()
    if result.nr is None:
        next_nr = 1
    else:
        next_nr = result.nr + 1
    return next_nr


def get_results(db_session):
    """ Return all entries of measurements with their average. The results are grouped by their name. """
    return db_session.query(TestRun.name,
                            func.count(TestRun.execution_time).label('count'),
                            func.avg(TestRun.execution_time).label('average')
                            ).group_by(TestRun.name).order_by(desc('count')).all()


def get_res_current(db_session, version):
    """ Return entries of measurements with their average from the current project version only. """
    return db_session.query(TestRun.name,
                            func.count(TestRun.execution_time).label('count'),
                            func.avg(TestRun.execution_time).label('average')) \
        .filter(TestRun.version == version) \
        .group_by(TestRun.name).order_by(desc('count')).all()


def get_line_results(db_session):
    return db_session.query(TestRun.version,
                            func.avg(TestRun.execution_time).label('avg'),
                            func.min(TestRun.execution_time).label('min'),
                            func.max(TestRun.execution_time).label('max'),
                            func.count(TestRun.execution_time).label('count')
                            ).group_by(TestRun.version).order_by(desc('count')).all()


def get_suites(db_session):
    return db_session.query(TestRun.suite).group_by(TestRun.suite).all()


def get_measurements(db_session, suite):
    """Return all measurements for some Travis build. Used for creating a box plot. """
    return db_session.query(TestRun).filter(TestRun.suite == suite).all()


def get_test_measurements(db_session, name, suite):
    """Return all measurements for some test of some Travis build. Used for creating a box plot. """
    return db_session.query(TestRun).filter(TestRun.name == name, TestRun.suite == suite).all()



import configparser
import os

from flask_monitoringdashboard.core.config.parser import parse_string, parse_version, parse_bool, parse_literal


class Config(object):
    """
        The settings can be changed by setting up a config file. For an example of a config file, see
         config.cfg in the main-directory. 
    """

    def __init__(self):
        """
            Sets the default values for the project
        """
        self.version = '1.0'
        self.link = 'dashboard'
        self.database_name = 'sqlite:///flask_monitoringdashboard.db'
        self.test_dir = None
        self.username = 'admin'
        self.password = 'admin'
        self.guest_username = 'guest'
        self.guest_password = ['guest_password']
        self.outlier_detection_constant = 2.5
        self.colors = {}
        self.security_token = 'cc83733cb0af8b884ff6577086b87909'
        self.outliers_enabled = True

        # define a custom function to retrieve the session_id or username
        self.group_by = None

    def init_from(self, file=None, envvar=None):
        """
            The config_file must at least contains the following variables in section 'dashboard':
            APP_VERSION: the version of the app that you use. Updating the version helps in 
                showing differences in execution times of a function over a period of time.
            CUSTOM_LINK: The dashboard can be visited at localhost:5000/{{CUSTOM_LINK}}.
            DATABASE: Suppose you have multiple projects where you're working on and want to 
                separate the results. Then you can specify different database_names, such that the 
                result of each project is stored in its own database.
            
            Since updating the version in the config-file when updating code isn't very useful, it
            is a better idea to provide the location of the git-folder. From the git-folder. The 
            version automatically retrieved by reading the commit-id (hashed value):
            GIT = If you're using git, then it is easier to set the location to the .git-folder, 
                The location is relative to the config-file.

            USERNAME: for logging into the dashboard, a username and password is required. The
                username can be set using this variable.
            PASSWORD: same as for the username, but this is the password variable.
            GUEST_USERNAME: A guest can only see the results, but cannot configure/download data.
            GUEST_PASSWORD: A guest can only see the results, but cannot configure/download data.

            OUTLIER_DETECTION_CONSTANT: When the execution time is more than this constant *
                average, extra information is logged into the database. A default value for this
                variable is 2.5, but can be changed in the config-file.

            SECURITY_TOKEN: Used for getting the data in /get_json_data/<security_token>

            OUTLIERS_ENABLED: Whether you want the Dashboard to collect extra information about outliers.

            :param file: a string pointing to the location of the config-file
            :param envvar: a string specifying which environment variable holds the config file location
        """

        if envvar:
            file = os.getenv(envvar)
        if not file:
            # Travis does not need a config file.
            if '/home/travis/build/' in os.getcwd():
                return
            print("No configuration file specified. Please do so.")
            return

        parser = configparser.RawConfigParser()
        try:
            parser.read(file)
            parse_version(parser, self.version)

            self.link = parse_string(parser, 'CUSTOM_LINK', self.link)
            self.database_name = parse_string(parser, 'DATABASE', self.database_name)
            self.test_dir = parse_string(parser, 'TEST_DIR', self.test_dir)
            self.security_token = parse_string(parser, 'SECURITY_TOKEN', self.security_token)
            self.outliers_enabled = parse_bool(parser, 'OUTLIERS_ENABLED', self.outliers_enabled)
            self.colors = parse_literal(parser, 'COLORS', self.colors)
            self.outlier_detection_constant = parse_literal(parser, 'OUTlIER_DETECTION_CONSTANT',
                                                            self.outlier_detection_constant)
            self.username = parse_string(parser, 'USERNAME', self.username)
            self.password = parse_string(parser, 'PASSWORD', self.password)
            self.guest_username = parse_string(parser, 'GUEST_USERNAME', self.guest_username)
            self.guest_password = parse_literal(parser, 'GUEST_PASSWORD', self.guest_password)
        except configparser.Error:
            raise



"""
    This file contains all unit tests for the endpoint-table in the database. (Corresponding to the file:
    'flask_monitoringdashboard/database/endpoint.py')
    See __init__.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, EXECUTION_TIMES, NAME, \
    GROUP_BY, IP


class TestEndpoint(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_get_monitor_rule(self):
        """
            Test wheter the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_monitor_rule
        from flask_monitoringdashboard import config
        with session_scope() as db_session:
            rule = get_monitor_rule(db_session, NAME)
        self.assertEqual(rule.endpoint, NAME)
        self.assertTrue(rule.monitor)
        self.assertEqual(rule.version_added, config.version)

    def test_update_monitor_rule(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_monitor_rule, update_monitor_rule
        with session_scope() as db_session:
            current_value = get_monitor_rule(db_session, NAME).monitor
            new_value = not current_value
            update_monitor_rule(db_session, NAME, new_value)
            self.assertEqual(get_monitor_rule(db_session, NAME).monitor, new_value)

    def test_get_all_measurement_per_column(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_all_measurement_per_column
        from flask_monitoringdashboard.database import FunctionCall
        from flask_monitoringdashboard import config
        with session_scope() as db_session:
            result = get_all_measurement_per_column(db_session, endpoint=NAME, column=FunctionCall.ip, value=IP)
            self.assertEqual(len(result), len(EXECUTION_TIMES))
            for row in result:
                self.assertIn(row.execution_time, EXECUTION_TIMES)
                self.assertEqual(row.version, config.version)
                self.assertEqual(row.endpoint, NAME)
                self.assertEqual(row.group_by, GROUP_BY)
                self.assertEqual(row.ip, IP)

    def test_update_last_accessed(self):
        """
            Test whether the function returns the right values.
        """
        import datetime
        time = datetime.datetime.now()
        from flask_monitoringdashboard.database.endpoint import update_last_accessed, get_last_accessed_times
        with session_scope() as db_session:
            update_last_accessed(db_session, NAME, time)
            result = get_last_accessed_times(db_session, NAME)
            self.assertEqual(result, time)



import ast

import numpy as np
from flask import url_for
from werkzeug.routing import BuildError

from flask_monitoringdashboard import config
from flask_monitoringdashboard.database.count import count_requests
from flask_monitoringdashboard.database.endpoint import get_monitor_rule
from flask_monitoringdashboard.database.function_calls import get_date_of_first_request


def get_endpoint_details(db_session, endpoint):
    """ Return details about an endpoint"""
    return {
        'endpoint': endpoint,
        'rule': get_monitor_rule(db_session, endpoint),
        'url': get_url(endpoint),
        'total_hits': count_requests(db_session, endpoint)
    }


def get_details(db_session):
    """ Return details about the deployment """
    from constants import VERSION

    return {
        'link': config.link,
        'dashboard-version': VERSION,
        'config-version': config.version,
        'first-request': get_date_of_first_request(db_session)
    }


def formatter(ms):
    """
    formats the ms into seconds and ms
    :param ms: the number of ms
    :return: a string representing the same amount, but now represented in seconds and ms.
    """
    sec = int(ms) // 1000
    ms = int(ms) % 1000
    if sec == 0:
        return '{0}ms'.format(ms)
    return '{0}.{1}s'.format(sec, ms)


def get_url(end):
    """
    Returns the URL if possible.
    URL's that require additional arguments, like /static/<file> cannot be retrieved.
    :param end: the endpoint for the url.
    :return: the url_for(end) or None,
    """
    try:
        return url_for(end)
    except BuildError:
        return None


def simplify(values, n=5):
    """
    Simplify a list of values. It returns a list that is representative for the input
    :param values: list of values
    :param n: length of the returned list
    :return: list with n values: min, q1, median, q3, max
    """
    return [np.percentile(values, i * 100 // (n - 1)) for i in range(n)]


def get_mean_cpu(cpu_percentages):
    """
    Returns a list containing mean CPU percentages per core for all given CPU percentages.
    :param cpu_percentages: list of CPU percentages
    """
    if not cpu_percentages:
        return None

    count = 0  # some outliers have no CPU info
    values = []  # list of lists that stores the CPU info

    for cpu in cpu_percentages:
        if not cpu:
            continue
        x = ast.literal_eval(cpu[0])
        values.append(x)
        count += 1

    sums = [sum(x) for x in zip(*values)]
    means = list(map(lambda x: round(x / count), sums))
    return means



import datetime

from flask import request

import flask_monitoringdashboard.views.export.csv
import flask_monitoringdashboard.views.export.json
from flask_monitoringdashboard import blueprint, config
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.tests import add_or_update_test, add_test_result, get_suite_nr
from flask_monitoringdashboard.database.tests_grouped import reset_tests_grouped, add_tests_grouped


@blueprint.route('/submit-test-results', methods=['POST'])
def submit_test_results():
    content = request.get_json()['test_runs']
    with session_scope() as db_session:
        suite = get_suite_nr(db_session)
        for result in content:
            time = datetime.datetime.strptime(result['time'], '%Y-%m-%d %H:%M:%S.%f')
            add_or_update_test(db_session, result['name'], time, result['successful'])
            add_test_result(db_session, result['name'], result['exec_time'], time, config.version, suite,
                            result['iter'])

        groups = request.get_json()['grouped_tests']
        if groups:
            reset_tests_grouped(db_session)
            add_tests_grouped(db_session, groups)

    return '', 204



import argparse
import csv
import os
import datetime
import time
from unittest import TestLoader

import requests

# Parsing the arguments.
parser = argparse.ArgumentParser(description='Collecting performance results from the unit tests of a project.')
parser.add_argument('--test_folder', dest='test_folder', required=True,
                    help='folder in which the unit tests can be found (example: ./tests)')
parser.add_argument('--times', dest='times', default=5,
                    help='number of times to execute every unit test (default: 5)')
parser.add_argument('--url', dest='url', default=None,
                    help='url of the Dashboard to submit the performance results to')
args = parser.parse_args()
print('Starting the collection of performance results with the following settings:')
print('  - folder containing unit tests: ', args.test_folder)
print('  - number of times to run tests: ', args.times)
print('  - url to submit the results to: ', args.url)
if not args.url:
    print('The performance results will not be submitted.')

# Initialize result dictionary and logs.
data = {'test_runs': [], 'grouped_tests': []}
home = os.path.expanduser("~")
log = open(home + '/endpoint_hits.log', 'w')
log.write('"time","endpoint"\n')
log.close()
log = open(home + '/test_runs.log', 'w')
log.write('"start_time","stop_time","test_name"\n')

# Find the tests and execute them the specified number of times.
# Add the performance results to the result dictionary.
suites = TestLoader().discover(args.test_folder, pattern="*test*.py")
for iteration in range(int(args.times)):
    for suite in suites:
        for case in suite:
            for test in case:
                test_result = None
                start_time_stamp = str(datetime.datetime.now())
                time_before = time.time()
                test_result = test.run(test_result)
                time_after = time.time()
                end_time_stamp = str(datetime.datetime.now())
                log.write('"{}","{}","{}"\n'.format(start_time_stamp, end_time_stamp, str(test)))
                execution_time = (time_after - time_before) * 1000
                data['test_runs'].append(
                    {'name': str(test), 'exec_time': execution_time, 'time': str(datetime.datetime.now()),
                     'successful': test_result.wasSuccessful(), 'iter': iteration + 1})
log.close()

# Read and parse the log containing the test runs into an array for processing.
test_runs = []
with open(home + '/test_runs.log') as log:
    reader = csv.DictReader(log)
    for row in reader:
        test_runs.append([datetime.datetime.strptime(row["start_time"], "%Y-%m-%d %H:%M:%S.%f"),
                          datetime.datetime.strptime(row["stop_time"], "%Y-%m-%d %H:%M:%S.%f"),
                          row['test_name']])

# Read and parse the log containing the endpoint hits into an array for processing.
endpoint_hits = []
with open(home + '/endpoint_hits.log') as log:
    reader = csv.DictReader(log)
    for row in reader:
        endpoint_hits.append([datetime.datetime.strptime(row["time"], "%Y-%m-%d %H:%M:%S.%f"),
                              row['endpoint']])

# Analyze the two arrays to find out which endpoints were hit by which unit tests.
# Add the endpoint_name/test_name combination to the result dictionary.
for endpoint_hit in endpoint_hits:
    for test_run in test_runs:
        if test_run[0] <= endpoint_hit[0] <= test_run[1]:
            if {'endpoint': endpoint_hit[1], 'test_name': test_run[2]} not in data['grouped_tests']:
                data['grouped_tests'].append({'endpoint': endpoint_hit[1], 'test_name': test_run[2]})
            break

# Send test results and endpoint_name/test_name combinations to the Dashboard if specified.
if args.url:
    if args.url[-1] == '/':
        args.url += 'submit-test-results'
    else:
        args.url += '/submit-test-results'
    try:
        requests.post(args.url, json=data)
        print('Sent unit test results to the Dashboard at ', args.url)
    except Exception as e:
        print('Sending unit test results to the dashboard failed:\n{}'.format(e))



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.utils import simplify
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.plot import boxplot, get_layout, get_figure, get_margin
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.function_calls import get_data_per_version
from flask_monitoringdashboard.database.versions import get_date_first_request, get_versions

TITLE = 'Global execution time for every version'


@blueprint.route('/measurements/versions')
@secure
def page_boxplot_per_version():
    return render_template('fmd_dashboard/graph.html', graph=get_boxplot_per_version(), title=TITLE)


def get_boxplot_per_version():
    """
    Creates a graph with the execution times per version
    :return:
    """
    with session_scope() as db_session:
        versions = get_versions(db_session)

    data = []
    for version in versions:
        values = [c.execution_time for c in get_data_per_version(db_session, version)]
        name = "{} {}".format(version, get_date_first_request(db_session, version).strftime("%b %d %H:%M"))
        data.append(boxplot(name=name, values=simplify(values, 10), marker={'color': get_color(version)}))

    layout = get_layout(
        height=350 + 40 * len(versions),
        title=TITLE,
        xaxis={'title': 'Execution time (ms)'},
        yaxis={'type': 'category', 'autorange': 'reversed'},
        margin=get_margin()
    )
    return get_figure(layout=layout, data=data)



"""
    Contains all endpoints that present information about specific endpoints.
    The endpoints are the following:
    - heatmap/<end>: shows a heatmap of the usage for <end>
    - Time per version per user/<end>: shows an overview of the time per user per version for a given <end>
    - Time per version per ip/<end>: shows an overview of the time per user per ip for a given <end>
    - Time per version/<end>: shows an overview of all requests per version for a given <end>
    - Time per user/<end>: shows an overview of all requests per user for a given <end>
    - Outliers/<end>: shows information about requests that take too long.
"""

from flask_monitoringdashboard.views.details.heatmap import result_heatmap
from flask_monitoringdashboard.views.details.outliers import result_outliers
from flask_monitoringdashboard.views.details.time_user import result_time_per_user
from flask_monitoringdashboard.views.details.time_version import result_time_per_version
from flask_monitoringdashboard.views.details.version_ip import result_time_per_version_per_ip
from flask_monitoringdashboard.views.details.version_user import result_time_per_version_per_user



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.plot import get_layout, get_figure, boxplot, get_margin
from flask_monitoringdashboard.core.utils import simplify
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.function_calls import get_endpoints, get_data_per_endpoint

TITLE = 'Global execution time for every endpoint'


@blueprint.route('/measurements/endpoints')
@secure
def page_boxplot_per_endpoint():
    return render_template('fmd_dashboard/graph.html', graph=get_boxplot_per_endpoint(), title=TITLE)


def get_boxplot_per_endpoint():
    """
    Creates a graph with the execution times per endpoint
    :return:
    """
    with session_scope() as db_session:
        endpoints = get_endpoints(db_session)

        data = []
        for endpoint in endpoints:
            values = [c.execution_time for c in get_data_per_endpoint(db_session, endpoint)]
            if values:
                data.append(boxplot(simplify(values, 10), name=endpoint))

    layout = get_layout(
        height=350 + 40 * len(endpoints),
        title=TITLE,
        xaxis={'title': 'Execution time (ms)'},
        margin=get_margin()
    )
    return get_figure(layout, data)



"""
    This file contains all unit tests for the monitor-rules-table in the database. (Corresponding to the file:
    'flask_monitoringdashboard/database/tests_grouped.py')
    See __init__.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, TEST_NAMES

NAME2 = 'main2'
SUITE = 3


class TestDBTestsGrouped(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_reset_tests_grouped(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests_grouped import reset_tests_grouped, get_tests_grouped
        with session_scope() as db_session:
            self.assertEqual(len(get_tests_grouped(db_session)), len(TEST_NAMES))
            reset_tests_grouped(db_session)
            self.assertEqual(get_tests_grouped(db_session), [])

    def test_add_tests_grouped(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests_grouped import add_tests_grouped, get_tests_grouped
        json = [{'endpoint': 'endpoint', 'test_name': 'test_name'}]
        with session_scope() as db_session:
            self.assertEqual(len(get_tests_grouped(db_session)), len(TEST_NAMES))
            add_tests_grouped(db_session, json)
            self.assertEqual(len(get_tests_grouped(db_session)), len(TEST_NAMES)+1)

    def test_get_tests_grouped(self):
        """
            Test whether the function returns the right values.
        """
        self.test_reset_tests_grouped()  # can be replaced by test_add_test_result, since this function covers two tests



"""
Contains all functions that access a single endpoint
"""
import datetime

from sqlalchemy import func, desc
from sqlalchemy.orm.exc import NoResultFound

from flask_monitoringdashboard import config
from flask_monitoringdashboard.database import FunctionCall, MonitorRule


def get_num_requests(db_session, endpoint, start_date, end_date):
    """ Returns a list with all dates on which an endpoint is accessed.
        :param endpoint: if None, the result is the sum of all endpoints
        :param start_date: datetime.date object
        :param end_date: datetime.date object
    """
    query = db_session.query(func.strftime('%Y-%m-%d %H:00:00', FunctionCall.time).label('newTime'),
                             func.count(FunctionCall.execution_time).label('count'))
    if endpoint:
        query = query.filter(FunctionCall.endpoint == endpoint)
    result = query.filter(FunctionCall.time >= datetime.datetime.combine(start_date, datetime.time(0, 0, 0, 0))). \
        filter(FunctionCall.time <= datetime.datetime.combine(end_date, datetime.time(23, 59, 59))). \
        group_by('newTime').all()
    db_session.expunge_all()
    return result


def get_group_by_sorted(db_session, endpoint, limit=None):
    """
    Returns a list with the distinct group-by from a specific endpoint. The limit is used to filter the most used
    distinct
    :param endpoint: the endpoint to filter on
    :param limit: the number of
    :return: a list with the group_by as strings.
    """
    query = db_session.query(FunctionCall.group_by, func.count(FunctionCall.group_by)). \
        filter(FunctionCall.endpoint == endpoint).group_by(FunctionCall.group_by). \
        order_by(desc(func.count(FunctionCall.group_by)))
    if limit:
        query = query.limit(limit)
    result = query.all()
    db_session.expunge_all()
    return [r[0] for r in result]


def get_ip_sorted(db_session, endpoint, limit=None):
    """
    Returns a list with the distinct group-by from a specific endpoint. The limit is used to filter the most used
    distinct
    :param endpoint: the endpoint to filter on
    :param limit: the number of
    :return: a list with the group_by as strings.
    """
    query = db_session.query(FunctionCall.ip, func.count(FunctionCall.ip)). \
        filter(FunctionCall.endpoint == endpoint).group_by(FunctionCall.ip). \
        order_by(desc(func.count(FunctionCall.ip)))
    if limit:
        query = query.limit(limit)
    result = query.all()
    db_session.expunge_all()
    return [r[0] for r in result]


def get_monitor_rule(db_session, endpoint):
    """ Get the MonitorRule from a given endpoint. If no value is found, a new value 
    is added to the database and the function is called again (recursively). """
    try:
        result = db_session.query(MonitorRule). \
            filter(MonitorRule.endpoint == endpoint).one()
        # for using the result when the session is closed, use expunge
        db_session.expunge(result)
        return result
    except NoResultFound:
        db_session.add(
            MonitorRule(endpoint=endpoint, version_added=config.version, time_added=datetime.datetime.now()))

    # return new added row
    return get_monitor_rule(db_session, endpoint)


def update_monitor_rule(db_session, endpoint, value):
    """ Update the value of a specific monitor rule. """
    db_session.query(MonitorRule).filter(MonitorRule.endpoint == endpoint). \
        update({MonitorRule.monitor: value})


def get_all_measurement_per_column(db_session, endpoint, column, value):
    """Return all entries with measurements from a given endpoint for which the column has a specific value.
    Used for creating a box plot. """
    result = db_session.query(FunctionCall).filter(FunctionCall.endpoint == endpoint, column == value).all()
    db_session.expunge_all()
    return result


def get_last_accessed_times(db_session, endpoint):
    """ Returns a list of all endpoints and their last accessed time. """
    result = db_session.query(MonitorRule.last_accessed).filter(MonitorRule.endpoint == endpoint).first()
    db_session.expunge_all()
    if result:
        return result[0]
    return None


def update_last_accessed(db_session, endpoint, value):
    """ Updates the timestamp of last access of the endpoint. """
    db_session.query(MonitorRule).filter(MonitorRule.endpoint == endpoint). \
        update({MonitorRule.last_accessed: value})



"""
Contains all functions that access any functionCall-object
"""

import datetime
import time

from sqlalchemy import func, distinct

from flask_monitoringdashboard import config
from flask_monitoringdashboard.database import FunctionCall
from flask_monitoringdashboard.database.count import count_requests


def get_requests_per_day(db_session, endpoint, list_of_days):
    """
    :param endpoint: name of the endpoint
    :param list_of_days: a list with datetime.date objects
    :return: A list of the amount of requests per day for a specific endpoint.
    If no requests is made on that day, a value of 0 is returned.
    """
    result_list = []
    for day in list_of_days:
        result = db_session.query(FunctionCall.execution_time). \
            filter(FunctionCall.endpoint == endpoint). \
            filter(func.strftime('%Y-%m-%d', FunctionCall.time) == day.strftime('%Y-%m-%d')).count()
        result_list.append(result)
    return result_list


def add_function_call(db_session, execution_time, endpoint, ip):
    """ Add a measurement to the database. """
    db_session.add(FunctionCall(endpoint=endpoint, execution_time=execution_time, version=config.version, ip=ip))


def get_median(db_session, endpoint, *where):
    """ Return the median for a specific endpoint within a certain time interval.
    If date_from is not specified, all results are counted
    :param endpoint: name of the endpoint
    :param where: additional where clause
    """
    num_rows = count_requests(db_session, endpoint, *where)
    result = db_session.query(FunctionCall.execution_time). \
        filter(FunctionCall.endpoint == endpoint, *where). \
        order_by(FunctionCall.execution_time).limit(2 - num_rows % 2).offset((num_rows - 1) / 2).all()
    values = [r[0] for r in result]
    return sum(values) / max(1, len(values))


def get_data_between(db_session, time_from, time_to=None):
    """
        Returns all data in the FunctionCall table, for the export data option.
        This function returns all data after the time_from date.
    """
    query = db_session.query(FunctionCall).filter(FunctionCall.time > time_from)
    if time_to:
        query = query.filter(FunctionCall.time <= time_to)
    result = query.all()
    db_session.expunge_all()
    return result


def get_data(db_session):
    """
    Equivalent function to get_data_from, but returns all data.
    :return: all data from the database in the Endpoint-table.
    """
    return get_data_between(db_session, datetime.date(1970, 1, 1), datetime.datetime.utcnow())


def get_data_per_version(db_session, version):
    """ Returns all data in the FunctionCall table, grouped by their version. """
    result = db_session.query(FunctionCall.execution_time, FunctionCall.version). \
        filter(FunctionCall.version == version).all()
    db_session.expunge_all()
    return result


def get_data_per_endpoint(db_session, end):
    result = db_session.query(FunctionCall.execution_time, FunctionCall.endpoint). \
        filter(FunctionCall.endpoint == end).all()
    db_session.expunge_all()
    return result


def get_endpoints(db_session):
    """ Returns the name of all endpoints from the database """
    result = db_session.query(distinct(FunctionCall.endpoint)).order_by(FunctionCall.endpoint).all()
    db_session.expunge_all()
    return [r[0] for r in result]  # unpack tuple result


def get_date_of_first_request(db_session):
    """ return the date (as unix timestamp) of the first request """
    result = db_session.query(FunctionCall.time).first()
    if result:
        return int(time.mktime(result[0].timetuple()))
    return -1
